#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from enum import Enum
from pathlib import Path
from typing import Iterator, Protocol

from pandas import DataFrame

from synnax.io.meta import ChannelMeta


class ReaderType(Enum):
    Row = "row"
    Column = "column"


class Matcher(Protocol):
    @classmethod
    def extensions(cls) -> list[str]:
        """:returns: a list of file extensions that the reader can read. Sometimes the
        list is not exhaustive, and is mostly used for informational purposes. For
        accurate file extension matching, use the :meth:`match` method.
        """
        ...

    @classmethod
    def match(cls, path: Path) -> bool:
        """:returns: whether the reader can read the file at the given path."""
        ...


class Closer(Protocol):
    """Closer is a closable buffer"""

    def close(self):
        """Closes the buffer."""
        ...


class File(Matcher, Closer, Protocol):
    """File is the base protocol for all file protocols. It's used to provide common
    information and utilities for all file protocols."""

    def path(self) -> Path:
        """:returns: the path to the file."""
        ...


class BaseReader(File, Protocol):
    """The base reader protocol that all other reader protocols must implement.

    :param path: The path to the file to read.
    :param keys: A list of channel keys to read. If None, all channels are read.
    :param chunk_size: The number of rows to read at a time. If None, all rows are read.
    """

    def __init__(
        self,
        path: Path,
        keys: list[str] | None = None,
        chunk_size: int | None = None,
    ): ...

    def channels(self) -> list[ChannelMeta]:
        """:returns : a list of channel metadata for the file."""
        ...

    @classmethod
    def type(cls) -> ReaderType:
        """:returns : the type of reader."""
        ...

    def path(self) -> Path:
        """:returns: the path to the file."""
        ...

    def nsamples(self) -> int:
        """:returns: the number of samples in the file."""
        ...

    def seek_first(self):
        """Seeks the reader to the first  sampele in the file."""
        ...


class RowFileReader(BaseReader, Protocol):
    """Row readers implement a strategy that reads a file row.py by row.py. Because Synnax
    is optimized for ingesting data in a columnar format, Row readers should
    only be used when files cannot be read using a :class:`ColumnReader` strategy (e.g.
    csv files).
    """

    def set_chunk_size(self, chunk_size: int):
        """Set the chunk size for the reader. It's generally unsafe to assume the reader
        position after calling set_chunk_size, so it's recommended to call reset
        afterwards.
        """
        ...

    def read(self) -> DataFrame:
        """Read returns a dataframe with chunk size * number of columns samples. The
        returned dataframe is guaranteed to contain a column for all channels in the file.
        """
        ...

    def __iter__(self) -> Iterator[DataFrame]:
        """Iterates over the file, returning a dataframe with chunk size * number of
        columns samples. The returned dataframe is guaranteed to contain a column for all
        channels in the file.
        """
        ...


class ColumnFileReader(BaseReader, Protocol):
    """Column readers implement a strategy that reads a file column by column. Synnax
    is optimized for ingesting data in a columnar format, so Column readers should
    be used whenever possible.
    """

    def read(self, *keys: str) -> DataFrame:
        """Reads a dataframe with chunk size * number of columns samples. The returned
        dataframe contains columns for each key in keys.
        """
        ...


class DataFrameWriter(Closer, Protocol):
    """A protocol for writing dataframes to a buffer. This protocol is kept separate
    from file protocols because it's also possible to write dataframes to other sources,
    such as a Synnax cluster.
    """

    def write(self, df: DataFrame) -> None | bool:
        """Writes the given dataframe to the buffer."""
        ...


class FileWriter(File, DataFrameWriter, Protocol):
    """A file protocol for writing dataframes"""

    def __init__(
        self,
        path: Path,
    ):
        """Creates a new file writer.

        :param path: The path to the file to write.
        """
        ...



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from synnax.hardware import device
from synnax.hardware import rack
from synnax.hardware import task


class Client:
    devices: device.Client
    tasks: task.Client
    racks: rack.Client

    def __init__(
        self,
        devices: device.Client,
        racks: rack.Client,
        tasks: task.Client,
    ) -> None:
        self.tasks = tasks
        self.racks = racks
        self.devices = devices



#  Copyright 2024 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from typing import overload
from freighter import Payload, UnaryClient, send_required, Empty
from synnax.ontology.payload import ID, Resource, Relationship, CrudeID
from synnax.ontology.group import Client as GroupClient
from synnax.util.normalize import normalize
from pydantic import Field


class RetrieveReq(Payload):
    ids: list[ID]
    children: bool = False
    parents: bool = False
    include_schema: bool = False
    exclude_field_data: bool = False
    term: str | None = None
    limit: int | None = None
    offset: int | None = None
    types: list[str] | None = None


class AddChildrenReq(Payload):
    id: ID
    children: list[ID]


class RemoveChildrenReq(Payload):
    id: ID
    children: list[ID]


class MoveChildrenReq(Payload):
    from_: ID = Field(alias="from")
    to: ID
    children: list[ID]


_RETRIEVE_ENDPOINT = "/ontology/retrieve"
_ADD_CHILDREN_ENDPOINT = "/ontology/add-children"
_REMOVE_CHILDREN_ENDPOINT = "/ontology/remove-children"
_MOVE_CHILDREN_ENDPOINT = "/ontology/move-children"


class RetrieveRes(Payload):
    resources: list[Resource]


class Client:
    _client: UnaryClient
    groups: GroupClient

    def __init__(self, client: UnaryClient, groups: GroupClient) -> None:
        self._client = client
        self.groups = groups

    @overload
    def retrieve(
        self,
        id: CrudeID,
        *,
        children: bool = False,
        parents: bool = False,
        include_schema: bool = False,
        exclude_field_data: bool = False
    ) -> Resource: ...

    @overload
    def retrieve(
        self,
        ids: list[CrudeID],
        *,
        children: bool = False,
        parents: bool = False,
        include_schema: bool = False,
        exclude_field_data: bool = False
    ) -> list[Resource]: ...

    def retrieve(
        self,
        id: CrudeID | list[CrudeID],
        *,
        children: bool = False,
        parents: bool = False,
        include_schema: bool = False,
        exclude_field_data: bool = False
    ) -> Resource | list[Resource]:
        is_single = False
        if not isinstance(id, list):
            id = [id]
            is_single = True
        req = RetrieveReq(
            ids=[ID(i) for i in id],
            children=children,
            parents=parents,
            include_schema=include_schema,
            exclude_field_data=exclude_field_data,
        )
        res = send_required(self._client, _RETRIEVE_ENDPOINT, req, RetrieveRes)
        if is_single:
            return res.resources[0]
        return res.resources

    def retrieve_children(
        self,
        id: CrudeID | list[CrudeID],
    ) -> list[Resource]:
        return send_required(
            self._client,
            _RETRIEVE_ENDPOINT,
            RetrieveReq(ids=[ID(i) for i in normalize(id)], children=True),
            RetrieveRes,
        ).resources

    def retrieve_parents(
        self,
        id: CrudeID | list[CrudeID],
    ):
        return send_required(
            self._client,
            _RETRIEVE_ENDPOINT,
            RetrieveReq(ids=[ID(i) for i in normalize(id)], parents=True),
            RetrieveRes,
        ).resources

    def move_children(self, from_: CrudeID, to: CrudeID, *children: CrudeID) -> None:
        send_required(
            self._client,
            _MOVE_CHILDREN_ENDPOINT,
            MoveChildrenReq.parse_obj(
                {"from": ID(from_), "to": ID(to), "children": [ID(i) for i in children]}
            ),
            Empty,
        )

    def remove_children(self, id: CrudeID, *children: CrudeID) -> None:
        send_required(
            self._client,
            _REMOVE_CHILDREN_ENDPOINT,
            RemoveChildrenReq(id=ID(id), children=[ID(i) for i in children]),
            Empty,
        )

    def add_children(self, id: CrudeID, *children: CrudeID) -> None:
        send_required(
            self._client,
            _ADD_CHILDREN_ENDPOINT,
            AddChildrenReq(parent=ID(id), children=[ID(i) for i in children]),
            Empty,
        )



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from alamos import Instrumentation, trace
from freighter import (
    Payload,
    UnaryClient,
    send_required,
)
from synnax.channel.payload import (
    ChannelKeys,
    ChannelNames,
    ChannelParams,
    normalize_channel_params,
)
from synnax.telem import TimeRange


class _Request(Payload):
    keys: ChannelKeys | None
    names: ChannelNames | None
    bounds: TimeRange


class _Response(Payload): ...


_ENDPOINT = "/frame/delete"


class Deleter:
    """
    Deleter is used to delete a time range of telemetry from the data engine.
    """

    _client: UnaryClient
    instrumentation: Instrumentation

    def __init__(self, client: UnaryClient, instrumentation: Instrumentation) -> None:
        self._client = client
        self.instrumentation = instrumentation

    @trace("debug")
    def delete(
        self,
        channels: ChannelParams,
        tr: TimeRange,
    ) -> None:
        normal = normalize_channel_params(channels)
        req = _Request(
            **{
                normal.variant: normal.channels,
                "bounds": tr,
            }
        )
        send_required(self._client, _ENDPOINT, req, _Response)



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from functools import wraps
from multiprocessing import Pool
from typing import Callable

from synnax import framer
from synnax.channel.payload import ChannelKey, ChannelName, ChannelParams
from synnax.channel.retrieve import ChannelRetriever
from synnax.state import LatestState, State

_InternalHandler = Callable[[State], Callable[[LatestState], None] | None]


class Registry:
    __handlers: list[_InternalHandler]
    __channels: set[ChannelKey | ChannelName]
    __frame_client: framer.Client
    __channel_retriever: ChannelRetriever

    def __init__(self, frame_client: framer.Client, channels: ChannelRetriever):
        self.__handlers = list()
        self.__channels = set()
        self.__frame_client = frame_client
        self.__channel_retriever = channels

    def on(
        self,
        channels: ChannelParams,
        filter_f: Callable[[LatestState], bool],
    ) -> Callable[[Callable[[LatestState], None]], Callable[[], None] | None]:
        self.__channels.update(channels)

        def decorator(f: Callable[[LatestState], None]) -> None:
            @wraps(f)
            def wrapper(state: State) -> Callable[[], None] | None:
                if filter_f(LatestState(state)):
                    return f
                return None

            self.__handlers.append(wrapper)
            return wrapper

        return decorator

    async def process(self) -> None:
        await Scheduler(
            channels=self.__channels,
            handlers=self.__handlers,
            frame_client=self.__frame_client,
            channel_retriever=self.__channel_retriever,
        ).start()


class Scheduler:
    __pool: Pool
    __streamer: framer.AsyncStreamer | None = None
    __handlers: list[_InternalHandler]
    __channels: ChannelParams
    __state: State
    __frame_client: framer.Client
    __channel_retriever: ChannelRetriever

    def __init__(
        self,
        channels: ChannelParams,
        handlers: list[_InternalHandler],
        frame_client: framer.Client,
        channel_retriever: ChannelRetriever,
    ):
        self.__frame_client = frame_client
        self.__channels = channels
        self.__handlers = handlers
        self.__channel_retriever = channel_retriever
        self.__state = State(channel_retriever)

    async def start(self):
        self.__streamer = await self.__frame_client.open_async_streamer(
            list(self.__channels)
        )
        async for frame in self.__streamer:
            self.__state.update(frame)
            for handler in self.__handlers:
                res = handler(self.__state)
                if res is not None:
                    res(LatestState(self.__state))

    def stop(self): ...



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from typing import overload, Iterator

from freighter import Payload
from pandas import DataFrame
from pydantic import Field

from synnax.channel.payload import (
    ChannelKey,
    ChannelKeys,
    ChannelName,
    ChannelNames,
)
from synnax.exceptions import ValidationError
from synnax.telem import Series, MultiSeries, TypedCrudeSeries, CrudeSeries


class FramePayload(Payload):
    keys: ChannelKeys
    series: list[Series]

    def __init__(
        self,
        keys: list[int] | None = None,
        series: list[Series] | None = None,
    ):
        # This is a workaround to allow for a None value to be
        # passed to the arrays field, but still have required
        # type hinting.
        if series is None:
            series = list()
        if keys is None:
            keys = list()
        super().__init__(series=series, keys=keys)


class Frame:
    """
    A frame is a collection of telemetry series mapped to particular channels. Frames
    can be keyed by channel name or channel key, but not both.
    """

    channels: list[ChannelKey | ChannelName]
    series: list[Series] = Field(default_factory=list)

    def __init__(
        self,
        channels: (
            ChannelKeys
            | ChannelNames
            | DataFrame
            | Frame
            | FramePayload
            | dict[ChannelKey, TypedCrudeSeries]
            | None
        ) = None,
        series: list[TypedCrudeSeries] | None = None,
    ):
        if isinstance(channels, Frame):
            self.channels = channels.channels
            self.series = channels.series
        elif isinstance(channels, FramePayload):
            self.channels = channels.keys
            self.series = channels.series
        elif isinstance(channels, DataFrame):
            self.channels = channels.columns.to_list()
            self.series = [Series(data=channels[k]) for k in self.channels]
        elif isinstance(channels, dict):
            self.channels = list(channels.keys())
            self.series = [Series(d) for d in channels.values()]
        elif (series is None or isinstance(series, list)) and (
            channels is None or isinstance(channels, list)
        ):
            self.series = list() if series is None else [Series(d) for d in series]
            self.channels = channels or list[ChannelKey]()
        else:
            raise ValueError(
                f"""
                [Frame] - invalid construction arguments. Received {channels}
                and {series}.
            """
            )

    def __str__(self) -> str:
        return self.to_df().__str__()

    @overload
    def append(self, col_or_frame: ChannelKey | ChannelName, series: Series) -> None:
        """Adds a new series to the frame for the given channel.
        :param col_or_frame: The channel key or name to append the series to.
        :param series: The series to append to the frame
        """
        ...

    @overload
    def append(self, col_or_frame: Frame) -> None:
        """Appends the given frame to the current frame, modifying the current frame in place.
        :param col_or_frame: The frame to append to the current frame.
        """
        ...

    def append(
        self,
        col_or_frame: ChannelKey | ChannelName | Frame,
        series: Series | None = None,
    ) -> None:
        """Appends a frame or series to the current frame. If a frame is provided, the
        series and channels from the frame are appended to the current frame. If a series
        is provided, the series is appended to the frame for the given channel.
        :param col_or_frame: The channel key or name to append the series to, or the frame
            to append to the current frame.
        :param series: The series to append to the frame, if col_or_frame is a channel key
            or name.
        """
        if isinstance(col_or_frame, Frame):
            self.series.extend(col_or_frame.series)  # type: ignore
            self.channels.extend(col_or_frame.channels)  # type: ignore
        else:
            self.series.append(series)
            self.channels.append(col_or_frame)  # type: ignore

    def items(
        self,
    ) -> Iterator[tuple[ChannelKey | ChannelName, Series]]:
        """
        Returns a generator of tuples containing the channel and series for each channel
        in the frame.
        """
        return zip(self.channels, self.series)  # type: ignore

    def __getitem__(self, key: ChannelKey | ChannelName | any) -> MultiSeries:
        if not isinstance(key, (ChannelKey, ChannelName)):
            return self.to_df()[key]
        indexes = [i for i, k in enumerate(self.channels) if k == key]
        return MultiSeries([self.series[i] for i in indexes])

    def get(
        self, key: ChannelKey | ChannelName, default: Series | None = None
    ) -> MultiSeries | None:
        """Gets the series for the given channel key or name. If the channel does not
        exist in the frame, returns the default value or None if no default is provided.
        :param key: The channel key or name to get the series for.
        :param default: The default value to return if the channel does not exist in the
        """
        try:
            return self[key]
        except ValueError:
            return default

    def to_payload(self):
        """Converts the frame to its payload representation for transport over the
        network. This method should typically only be used internally.
        :raises: ValidationError if the frame is keyed by channel name instead of key.
        """
        if not all(isinstance(k, ChannelKey) for k in self.channels):
            diff = [k for k in self.channels if not isinstance(k, ChannelKey)]
            raise ValidationError(
                f"""
            Cannot convert a frame that contains channel names to a payload.
            The following channels are invalid: {diff}
            """
            )
        return FramePayload(keys=self.channels, series=self.series)

    def to_df(self) -> DataFrame:
        """Converts the frame to a pandas DataFrame. Each column in the DataFrame
        corresponds to a channel in the frame.
        """
        base = dict()
        for k in set(self.channels):
            # Try to convert the value to a numpy array. If it fails (such as in the
            # case of strings or JSON objects), convert it to a primitive list instead.
            try:
                base[k] = self.get(k).__array__()
            except TypeError:
                base[k] = list(self.get(k))
        return DataFrame(base)

    def __contains__(self, key: ChannelKey | ChannelName) -> bool:
        return key in self.channels


CrudeFrame = Frame | FramePayload | dict[ChannelKey, CrudeSeries] | DataFrame



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from typing import Protocol

from alamos import NOOP, Instrumentation, trace
from freighter import Payload, UnaryClient, send_required

from synnax.channel.payload import (
    ChannelKey,
    ChannelKeys,
    ChannelName,
    ChannelNames,
    ChannelParams,
    ChannelPayload,
    normalize_channel_params,
)
from synnax.exceptions import NotFoundError


class _Request(Payload):
    names: list[str] | None = None
    keys: list[int] | None = None
    leaseholder: int | None = None


class _Response(Payload):
    channels: list[ChannelPayload] = list()
    not_found: list[str] | None = list()


class ChannelRetriever(Protocol):
    """Protocol for retrieving channel payloads from the cluster."""

    def retrieve(self, channels: ChannelParams) -> list[ChannelPayload]: ...

    def retrieve_one(self, param: ChannelKey | ChannelName) -> ChannelPayload: ...


_ENDPOINT = "/channel/retrieve"


class ClusterChannelRetriever:
    _client: UnaryClient
    instrumentation: Instrumentation

    def __init__(
        self,
        client: UnaryClient,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self._client = client
        self.instrumentation = instrumentation

    def _(self) -> ChannelRetriever:
        return self

    @trace("debug")
    def retrieve(self, channels: ChannelParams) -> list[ChannelPayload]:
        normal = normalize_channel_params(channels)
        if len(normal.channels) == 0:
            return list()
        req = _Request(**{normal.variant: normal.channels})
        return send_required(self._client, _ENDPOINT, req, _Response).channels

    @trace("debug")
    def retrieve_one(self, param: ChannelKey | ChannelName) -> ChannelPayload | None:
        req = _Request()
        if isinstance(param, ChannelKey):
            req.keys = [param]
        else:
            req.names = [param]
        res = send_required(self._client, _ENDPOINT, req, _Response)
        if len(res.channels) == 0:
            raise NotFoundError(f"Could not find channel matching {param}")
        return res.channels[0]


class CacheChannelRetriever:
    _retriever: ChannelRetriever
    _channels: dict[ChannelKey, ChannelPayload]
    _names_to_keys: dict[ChannelName, set[ChannelKey]]
    instrumentation: Instrumentation

    def __init__(
        self,
        retriever: ChannelRetriever,
        instrumentation: Instrumentation,
    ) -> None:
        self._channels = dict()
        self._names_to_keys = dict()
        self.instrumentation = instrumentation
        self._retriever = retriever

    def delete(self, keys: ChannelParams) -> None:
        normal = normalize_channel_params(keys)
        if normal.variant == "names":
            matches = {
                ch for ch in self._channels.values() if ch.name in normal.channels
            }
            for ch in matches:
                self._channels.pop(ch.key)
                self._names_to_keys.pop(ch.name)
        else:
            for key in normal.channels:
                channel = self._channels.get(key)
                if channel is not None:
                    self._channels.pop(key)
                    self._names_to_keys.pop(channel.name)

    def rename(self, keys: list[ChannelKey], names: list[ChannelName]) -> None:
        for key, name in zip(keys, names):
            channel = self._channels.get(key)
            if channel is None:
                continue
            self._channels.pop(key)
            existing_keys = self._names_to_keys.get(channel.name)
            if existing_keys is not None:
                existing_keys.remove(key)
            channel.name = name
            self._channels[channel.key] = channel
            existing_keys = self._names_to_keys.get(name)
            if existing_keys is None:
                self._names_to_keys[name] = {channel.key}
            else:
                existing_keys.add(channel.key)

    def _(self) -> ChannelRetriever:
        return self

    def _get(self, param: ChannelKey | ChannelName) -> list[ChannelPayload] | None:
        if isinstance(param, ChannelKey):
            ch = self._channels.get(param)
            return [ch] if ch is not None else None
        keys = self._names_to_keys.get(param, set())
        channels = list()
        for key in keys:
            ch = self._channels.get(key)
            if ch is not None:
                channels.append(ch)
        return None if len(channels) == 0 else channels

    def _get_one(self, param: ChannelKey | ChannelName) -> ChannelPayload | None:
        if isinstance(param, ChannelKey):
            return self._channels.get(param)
        keys = self._names_to_keys.get(param, None)
        if keys is None:
            return None
        return self._channels.get(next(iter(keys)))

    def set(self, channels: list[ChannelPayload]) -> None:
        for channel in channels:
            self._set_one(channel)

    def _set_one(self, channel: ChannelPayload) -> None:
        self._channels[channel.key] = channel
        keys = self._names_to_keys.get(channel.name)
        if keys is None:
            self._names_to_keys[channel.name] = {channel.key}
        else:
            keys.add(channel.key)

    @trace("debug")
    def retrieve(self, channels: ChannelParams) -> list[ChannelPayload]:
        normal = normalize_channel_params(channels)
        results = list()
        to_retrieve: ChannelKeys | ChannelNames = list()  # type: ignore
        for p in normal.channels:
            ch = self._get(p)
            if ch is None:
                to_retrieve.append(p)  # type: ignore
            else:
                results.extend(ch)

        if len(to_retrieve) == 0:
            return results

        retrieved = self._retriever.retrieve(to_retrieve)
        self.set(retrieved)
        results.extend(retrieved)
        return results

    def retrieve_one(self, param: ChannelKey | ChannelName) -> ChannelPayload | None:
        ch = self._get_one(param)
        if ch is not None:
            return ch
        retrieved = self._retriever.retrieve_one(param)
        if retrieved is not None:
            self._set_one(retrieved)
        return retrieved


def retrieve_required(
    r: ChannelRetriever, channels: ChannelParams
) -> list[ChannelPayload]:
    normal = normalize_channel_params(channels)
    results = r.retrieve(channels)
    not_found = list()
    for p in normal.channels:
        ch = next((c for c in results if c.key == p or c.name == p), None)
        if ch is None:
            not_found.append(p)
    if len(not_found) > 0:
        raise NotFoundError(f"Could not find channels: {not_found}")
    return results



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

import uuid
from datetime import datetime, timedelta, tzinfo, UTC
from math import trunc
from typing import ClassVar, Literal, TypeAlias, cast, get_args

import numpy as np
import pandas as pd
from numpy.typing import DTypeLike
from pydantic import BaseModel

from synnax.exceptions import ContiguityError


class TimeStamp(int):
    """TimeStamp represents a 64 bit nanosecond-precision UTC timestamp. The TimeStamp
    constructor accepts a variety of types and will attempt to convert them to a
    TimeStamp. The following types are supported:

    * TimeStamp - Returns a copy of the TimeStamp.
    * TimeSpan - Treats the TimeSpan as a duration since the Unix epoch in UTC.
    * pd.TimeStamp - Converts the pandas TimeStamp to a TimeStamp. If the timestamp is
        not timezone aware, it is assumed to be in the local timezone.
    * datetime - Converts the datetime to a TimeStamp.  If the datetime is not timezone
        aware, it is assumed to be in the local timezone.
    * timedelta - Treats the timedelta as a duration since the Unix epoch in UTC.
    * np.datetime64 - Treats the datetime64 as a duration since the Unix epoch in UTC.
    * int - Treats the int as a nanosecond duration since the Unix epoch in UTC.
    """

    def __new__(cls, value: CrudeTimeStamp):
        if isinstance(value, str):
            value = int(value)
        elif isinstance(value, TimeStamp):
            return value
        elif isinstance(value, TimeSpan):
            value = int(value)
        elif isinstance(value, pd.Timestamp):
            value = int(
                float(TimeSpan.SECOND) * value.to_pydatetime().astimezone().timestamp()
            )
        elif isinstance(value, datetime):
            value = int(float(TimeSpan.SECOND) * value.timestamp())
        elif isinstance(value, timedelta):
            value = int(float(TimeSpan.SECOND) * value.total_seconds())
        elif isinstance(value, np.datetime64):
            value = int(pd.Timestamp(value).value)
        elif isinstance(value, (int, float, np.integer, np.floating)):
            value = int(value)
        else:
            raise TypeError(f"Cannot convert {type(value)} to TimeStamp")

        return super().__new__(cls, value)

    @classmethod
    def __get_validators__(cls):
        """Implemented for pydantic validation"""
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """Implemented for pydantic validation"""
        # if its a string, cast the string to a number
        return cls(v)

    @classmethod
    def now(cls) -> TimeStamp:
        """:returns: the current time as a TimeStamp."""
        return TimeStamp(datetime.now())

    @classmethod
    def since(cls, ts: CrudeTimeStamp) -> TimeSpan:
        """:returns: a TimeSpan representing the amount of time elapsed since the given
        TimeStamp."""
        return TimeStamp.now().span(ts)

    def trunc(self, span: CrudeTimeSpan) -> TimeStamp:
        """Truncates the TimeSpan to the nearst integer multiple of the given span.

        For example,

        (TimeSpan.DAY + TimeSpan.HOUR + TimeSpan.SECOND).trunc(TimeSpan.DAY)

        would return

        TimeSpan.DAY

        :param span: The TimeSpan to truncate by.
        """
        return TimeStamp(TimeSpan(self).trunc(span))

    def span(self, other: CrudeTimeStamp) -> TimeSpan:
        """:returns: The TimeSpan between two timestamps. This span is guaranteed to be
        positive.
        """
        return TimeRange(self, other).make_valid().span

    def datetime(self, tz: tzinfo | None = None) -> datetime:
        """:returns: The TimeStamp represented as a timezone aware datetime object.
        :param tz: The timezone to use for the datetime. If not provided, the local
        timezone is used.
        """
        return datetime.fromtimestamp(self / TimeSpan.SECOND, UTC).astimezone(tz)

    def after(self, ts: CrudeTimeStamp) -> bool:
        """:returns: True if the TimeStamp is strictly after the given TimeStamp.
        :param ts: The TimeStamp to compare to.

        """
        return super().__gt__(TimeStamp(ts))

    def after_eq(self, ts: CrudeTimeStamp) -> bool:
        """:return: True if the TimeStamp is after or equal to the given TimeStamp.
        :param ts: The TimeStamp to compare to.
        """
        return super().__ge__(TimeStamp(ts))

    def before(self, ts: CrudeTimeStamp) -> bool:
        """:returns: True if the TimeStamp is strictly before the given TimeStamp.
        :param ts: The TimeStamp to compare to.
        """
        return super().__lt__(TimeStamp(ts))

    def before_eq(self, ts: CrudeTimeStamp) -> bool:
        """:returns: True if the TimeStamp is before or equal to the given TimeStamp.
        :param ts: The TimeStamp to compare to.
        """
        return super().__le__(TimeStamp(ts))

    def span_range(self, span: TimeSpan) -> TimeRange:
        """:returns: A TimeRange that spans the given TimeSpan
        :param span: The TimeSpan.
        """
        return TimeRange(self, self + span).make_valid()

    def range(self, ts: CrudeTimeStamp) -> TimeRange:
        """:return: A new TimeRange spanning the TimeStamp and provided TimeStamp. This
        TimeRange is guaranteed to be valid i.e. start before or equal to end.
        :param ts: The second time stamp
        """
        return TimeRange(self, TimeStamp(ts)).make_valid()

    def __add__(self, rhs: CrudeTimeStamp) -> TimeStamp:
        return TimeStamp(super().__add__(TimeStamp(rhs)))

    def __sub__(self, rhs: CrudeTimeStamp) -> TimeStamp:
        return TimeStamp(super().__sub__(TimeStamp(rhs)))

    def __lt__(self, rhs: CrudeTimeStamp) -> bool:
        return self.before(rhs)

    def __le__(self, rhs: CrudeTimeStamp) -> bool:
        return self.before_eq(rhs)

    def __ge__(self, rhs: CrudeTimeStamp) -> bool:
        return self.after_eq(rhs)

    def __gt__(self, rhs: CrudeTimeStamp) -> bool:
        return self.after(rhs)

    def __eq__(self, rhs: object) -> bool:
        if not isinstance(rhs, get_args(CrudeTimeStamp)):
            return False
        return super().__eq__(TimeStamp(cast(CrudeTimeStamp, rhs)))

    def __str__(self) -> str:
        return self.datetime().isoformat()

    MIN: TimeStamp
    """The minimum possible value of a TimeStamp"""
    MAX: TimeStamp
    """The maximum possible value of a TimeStamp"""
    ZERO: TimeStamp
    """The zero value of a TimeStamp"""


class TimeSpan(int):
    """TimeSpan represents a 64 bit nanosecond-precision duration. The TimeSpan
    constructor accepts a variety of different types and will attempt to convert them
    to a TimeSpan. The supported types are parsed as follows:

    * int - The number of nanoseconds.
    * np.int64 - The number of nanoseconds.
    * TimeSpan - Creates a copy of the TimeSpan.
    * TimeStamp - The difference between the TimeStamp and the Unix epoch
    * timedelta - The duration of the timedelta.
    * pands.timedelta - The duration of the timedelta.
    * np.timedelta64 - The duration of the timedelta64.
    """

    def __new__(cls, value: CrudeTimeSpan):
        if isinstance(value, str):
            value = int(value)
        if isinstance(value, timedelta):
            value = int(float(TimeSpan.SECOND) * value.total_seconds())
        elif isinstance(value, np.timedelta64):
            value = int(float(TimeSpan.SECOND) * pd.Timedelta(value).total_seconds())
        elif not isinstance(value, (int, float, np.floating, np.integer)):
            raise TypeError(f"Cannot convert {type(value)} to TimeSpan")
        return super().__new__(cls, value)

    @classmethod
    def __get_validators__(cls):
        """Implemented for pydantic validation. Should not be used externally."""
        yield cls.validate

    @classmethod
    def validate(cls, value):
        """Implemented for pydantic validation. Should not be used externally."""
        return cls(value)

    @classmethod
    def since(cls, ts: CrudeTimeStamp) -> TimeSpan:
        """:returns: a TimeStamp representing the amount of time elapsed since the given
        TimeStamp."""
        return TimeStamp.now().span(ts)

    def __repr__(self) -> str:
        return f"TimeSpan({super().__repr__()})"

    def __str__(self) -> str:
        tot_days = self.trunc(TimeSpan.DAY)
        tot_hours = self.trunc(TimeSpan.HOUR)
        tot_minutes = self.trunc(TimeSpan.MINUTE)
        tot_seconds = self.trunc(TimeSpan.SECOND)
        tot_millis = self.trunc(TimeSpan.MILLISECOND)
        tot_micros = self.trunc(TimeSpan.MICROSECOND)
        tot_nanos = self.trunc(TimeSpan.NANOSECOND)
        days = tot_days
        hours = tot_hours - tot_days
        minutes = tot_minutes - tot_hours
        seconds = tot_seconds - tot_minutes
        millis = tot_millis - tot_seconds
        micros = tot_micros - tot_millis
        nanos = tot_nanos - tot_micros

        v = ""
        if days != 0:
            v += f"{days.days_int}{TimeSpan.DAY_UNITS} "
        if hours != 0:
            v += f"{hours.hours_int}{TimeSpan.HOUR_UNITS} "
        if minutes != 0:
            v += f"{minutes.minutes_int}{TimeSpan.MINUTE_UNITS} "
        if seconds != 0:
            v += f"{seconds.seconds_int}{TimeSpan.SECOND_UNITS} "
        if millis != 0:
            v += f"{millis.milliseconds_int}{TimeSpan.MILLISECOND_UNITS} "
        if micros != 0:
            v += f"{micros.microseconds_int}{TimeSpan.MICROSECOND_UNITS} "
        if nanos != 0 or len(v) == 0:
            v += f"{nanos.nanoseconds}{TimeSpan.NANOSECOND_UNITS}"
        return v.strip()

    @staticmethod
    def from_seconds(span: CrudeTimeSpan) -> TimeSpan:
        """An alternative constructor for TimeSpan that parses raw numeric values as
        seconds instead of nanoseconds.
        """
        if isinstance(span, TimeSpan):
            return span
        if isinstance(span, Rate):
            return span.period
        if isinstance(span, (float, int)):
            return TimeSpan.SECOND * span
        return TimeSpan(span)

    @staticmethod
    def to_seconds(dur: CrudeTimeSpan | None) -> float | None:
        if dur is None:
            return None
        return TimeSpan.from_seconds(dur).seconds

    @property
    def rate(self) -> Rate:
        """:returns: The rate of the TimeSpan as a Rate."""
        return Rate(1 / self.seconds)

    @property
    def days(self) -> float:
        """:returns: The decimal number of days in the TimeSpan, including fractional
        days.
        """
        return float(self) / float(TimeSpan.DAY)

    @property
    def days_int(self) -> int:
        """:returns:  The integer number of days in the TimeSpan, NOT including
        fractional days.
        """
        return int(self / TimeSpan.DAY)

    @property
    def hours(self) -> float:
        """:returns: The decimal number of hours in the TimeSpan, including fractional
        hours.
        """
        return float(self) / float(TimeSpan.HOUR)

    @property
    def hours_int(self) -> int:
        """:returns: The integer number of hours in the TimeSpan, NOT including
        fractional hours.
        """
        return int(self / TimeSpan.HOUR)

    @property
    def minutes(self) -> float:
        """:returns: The decimal number of minutes in the TimeSpan, including fractional
        minutes.
        """
        return float(self) / float(TimeSpan.MINUTE)

    @property
    def minutes_int(self) -> int:
        """:returns: The integer number of minutes in the TimeSpan, NOT including
        fractional minutes.
        """
        return int(self / TimeSpan.MINUTE)

    @property
    def seconds(self) -> float:
        """:returns: The decimal number of seconds in the TimeSpan, including fractional
        seconds.
        """
        return float(self) / float(TimeSpan.SECOND)

    @property
    def seconds_int(self) -> int:
        """:returns: The integer number of seconds in the TimeSpan, NOT including
        fractional seconds.
        """
        return int(self / TimeSpan.SECOND)

    @property
    def milliseconds(self) -> float:
        """:returns: The decimal number of milliseconds in the TimeSpan, including the
        fractional milliseconds.
        """
        return float(self) / float(TimeSpan.MILLISECOND)

    @property
    def milliseconds_int(self) -> int:
        """:returns: The integer number of milliseconds in the TimeSpan, NOT including
        fractional milliseconds.
        """
        return int(self / TimeSpan.MILLISECOND)

    @property
    def microseconds(self) -> float:
        """:returns: The decimal number of microseconds in the TimeSpan, including
        the fractional microseconds.
        """
        return float(self) / float(TimeSpan.MICROSECOND)

    @property
    def microseconds_int(self) -> int:
        """:returns: The integer number of microseconds in the TimeSpan, NOT including
        fractional microseconds.
        """
        return int(self / TimeSpan.MICROSECOND)

    @property
    def nanoseconds(self) -> int:
        """:returns: The integer number of nanoseconds in the TimeSpan."""
        return int(self / TimeSpan.NANOSECOND)

    def trunc(self, span: CrudeTimeSpan) -> TimeSpan:
        """Truncates the TimeSpan to the nearest integer multiple of the given span.

        For example,

        (TimeSpan.DAY + TimeSpan.HOUR + TimeSpan.SECOND).trunc(TimeSpan.DAY)

        would return

        TimeSpan.DAY

        :param span: The TimeSpan to truncate by.
        """
        return TimeSpan(trunc(self / span) * span)

    @property
    def timedelta(self) -> timedelta:
        """:returns: The TimeSpan represented as a datetime.timedelta."""
        return timedelta(seconds=self.seconds)

    def __add__(self, rhs: CrudeTimeSpan) -> TimeSpan:
        return TimeSpan(super().__add__(TimeSpan(rhs)))

    def __sub__(self, rhs: CrudeTimeSpan) -> TimeSpan:
        return TimeSpan(super().__sub__(TimeSpan(rhs)))

    def __mul__(self, rhs: CrudeTimeSpan) -> TimeSpan:
        return TimeSpan(self.nanoseconds * rhs)

    def __truediv__(self, rhs: CrudeTimeSpan) -> TimeSpan:
        return TimeSpan(super().__truediv__(TimeSpan(rhs)))

    def __mod__(self, rhs: CrudeTimeSpan) -> TimeSpan:
        return TimeSpan(super().__mod__(TimeSpan(rhs)))

    def __rmul__(self, rhs: CrudeTimeSpan) -> TimeSpan:
        return self.__mul__(rhs)

    def __gt__(self, rhs: CrudeTimeSpan) -> bool:
        return super().__gt__(TimeSpan(rhs))

    def __ge__(self, rhs: CrudeTimeSpan) -> bool:
        return super().__ge__(TimeSpan(rhs))

    def __lt__(self, rhs: CrudeTimeSpan) -> bool:
        return super().__lt__(TimeSpan(rhs))

    def __le__(self, rhs: CrudeTimeSpan) -> bool:
        return super().__le__(TimeSpan(rhs))

    def __eq__(self, rhs: object) -> bool:
        if not isinstance(rhs, get_args(CrudeTimeSpan)):
            return NotImplemented
        return super().__eq__(int(TimeSpan(cast(CrudeTimeSpan, rhs))))

    NANOSECOND: TimeSpan
    """A nanosecond."""
    NANOSECOND_UNITS: TimeSpanUnits
    """The unit string for nanoseconds: 'ns'."""
    MICROSECOND: TimeSpan
    """A microsecond."""
    MICROSECOND_UNITS: TimeSpanUnits
    """The unit string for microseconds: 'us'."""
    MILLISECOND: TimeSpan
    """A millisecond."""
    MILLISECOND_UNITS: TimeSpanUnits
    """The unit string for milliseconds: 'ms'."""
    SECOND: TimeSpan
    """A second."""
    SECOND_UNITS: TimeSpanUnits
    """The unit string for seconds: 's'."""
    MINUTE: TimeSpan
    """A minute."""
    MINUTE_UNITS: TimeSpanUnits
    """The unit string for minutes: 'm'."""
    HOUR: TimeSpan
    """An hour."""
    HOUR_UNITS: TimeSpanUnits
    """The unit string for hours: 'h'."""
    DAY: TimeSpan
    """A day."""
    DAY_UNITS: TimeSpanUnits
    """The unit string for days: 'd'."""
    MAX: TimeSpan
    """The maximum possible value for a TimeSpan"""
    MIN: TimeSpan
    """The minimum possible value for a TimeSpan"""
    ZERO: TimeSpan
    """A dictionary mapping unit string to its corresponding TimeSan"""
    UNITS: dict[TimeSpanUnits, TimeSpan]


TimeSpan.NANOSECOND = TimeSpan(1)
TimeSpan.NANOSECOND_UNITS = "ns"
TimeSpan.MICROSECOND = TimeSpan(1000) * TimeSpan.NANOSECOND
TimeSpan.MICROSECOND_UNITS = "us"
TimeSpan.MILLISECOND = TimeSpan(1000) * TimeSpan.MICROSECOND
TimeSpan.MILLISECOND_UNITS = "ms"
TimeSpan.SECOND = TimeSpan(1000) * TimeSpan.MILLISECOND
TimeSpan.SECOND_UNITS = "s"
TimeSpan.MINUTE = TimeSpan(60) * TimeSpan.SECOND
TimeSpan.MINUTE_UNITS = "m"
TimeSpan.HOUR = TimeSpan(60) * TimeSpan.MINUTE
TimeSpan.HOUR_UNITS = "h"
TimeSpan.DAY = TimeSpan.HOUR * 24
TimeSpan.DAY_UNITS = "d"
TimeSpan.MAX = TimeSpan(0xFFFFFFFFFFFFFFFF)
TimeSpan.MIN = TimeSpan(0)
TimeSpan.ZERO = TimeSpan(0)
TimeSpan.UNITS = {
    TimeSpan.NANOSECOND_UNITS: TimeSpan.NANOSECOND,
    TimeSpan.MICROSECOND_UNITS: TimeSpan.MICROSECOND,
    TimeSpan.MILLISECOND_UNITS: TimeSpan.MILLISECOND,
    TimeSpan.SECOND_UNITS: TimeSpan.SECOND,
    TimeSpan.MINUTE_UNITS: TimeSpan.MINUTE,
    TimeSpan.HOUR_UNITS: TimeSpan.HOUR,
    TimeSpan.DAY_UNITS: TimeSpan.DAY,
}
TimeStamp.MIN = TimeStamp(0)
TimeStamp.ZERO = TimeStamp.MIN
TimeStamp.MAX = TimeStamp(2**63 - 1)

TimeSpanUnits = Literal["ns", "us", "ms", "s", "m", "h", "d", "iso"]


def convert_time_units(data: np.ndarray, _from: TimeSpanUnits, to: TimeSpanUnits):
    """Converts the data from one time unit to another.

    :param data: the numpy array to convert.
    :param _from: the units of the data
    :param to: the units to convert to
    :return: the data in the new units
    """
    if _from == "iso":
        data = np.array([datetime.fromisoformat(v).timestamp() for v in data])
        f = TimeSpan.SECOND
    else:
        f = TimeSpan.UNITS.get(_from, TimeSpan.ZERO)
    if f == 0:
        raise ValueError(f"Invalid input time unit {_from}")
    t = TimeSpan.UNITS.get(to, TimeSpan.ZERO)
    if t == 0:
        raise ValueError(f"Invalid output time unit {to}")

    converted = data * f / t
    if to == TimeSpan.SECOND_UNITS:
        return converted.astype(np.float64)
    if to == TimeSpan.NANOSECOND_UNITS:
        return converted.astype(np.int64)
    return converted


class Rate(float):
    """Rate represents a data rate measured in Hz."""

    def __new__(cls, value: CrudeRate):
        if isinstance(value, float):
            return super().__new__(cls, value)
        if isinstance(value, TimeSpan):
            value = 1 / value.seconds
        elif isinstance(value, int):
            value = float(value)
        else:
            raise TypeError(f"Cannot convert {type(value)} to Rate")
        return super().__new__(cls, value)

    @classmethod
    def __get_validators__(cls):
        """Implemented for pydantic validation. Should not be used externally."""
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """Implemented for pydantic validation. Should not be used externally."""
        return cls(v)

    @property
    def period(self) -> TimeSpan:
        """:returns: the period of the rate as a TimeSpan"""
        return TimeSpan(int(1 / self * float(TimeSpan.SECOND)))

    def sample_count(self, time_span: CrudeTimeSpan) -> int:
        """:returns: the number of samples in the given TimeSpan at this rate"""
        return int(TimeSpan(time_span).seconds * self)

    def byte_size(self, time_span: CrudeTimeSpan, density: Density) -> Size:
        """:returns: the amount of bytes occupied by the given TimeSpan at this rate
        and the given sample density.

        :param time_span: TimeSpan - A crude TimeSpan representing the size of the
        data set in time.
        :param density: Density - The sample density of the data.
        """
        return Size(self.sample_count(time_span) * int(density))

    def span(self, sample_count: int) -> TimeSpan:
        """:returns: the TimeSpan that corresponds to the given number of samples at
        this rate."""
        return self.period * sample_count

    def size_span(self, size: Size, density: Density) -> TimeSpan:
        """:returns: the TimeSpan that corresponds to the given number of bytes at this
        rate and the given sample density."""
        if size % density != 0:
            raise ContiguityError(f"Size {size} is not a multiple of density {density}")
        return self.span(int(size / density))

    def __str__(self):
        if self < 1:
            return f"{self.period} per cycle"
        return f"{round(self, 2)} Hz"

    def __repr__(self):
        return f"Rate({super().__repr__()} Hz)"

    def __mul__(self, rhs: CrudeRate) -> Rate:
        return Rate(super().__mul__(Rate(rhs)))

    def __rmul__(self, other) -> Rate:
        return self.__mul__(other)

    HZ: Rate
    """One Hz."""
    KHZ: Rate
    """One kHz"""
    MHZ: Rate
    """One MHz"""


Rate.HZ = Rate(1)
Rate.KHZ = Rate(1000) * Rate.HZ
Rate.MHZ = Rate(1000) * Rate.KHZ


class TimeRange(BaseModel):
    """TimeRange is a range of time marked by a start and end TimeStamp. A TimeRange
    is start inclusive and end exclusive.
    """

    start: TimeStamp
    """The starting TimeStamp of the TimeRange.

    Note that this value is not guaranteed to be before or equal to the ending value.
    To ensure that this is the case, call TimeRange.make_valid().

    In most cases, operations should treat start as inclusive.
    """
    end: TimeStamp
    """The ending TimeStamp of the TimeRange.

    Note that this value is not guaranteed to be after or equal to the ending value. To
    ensure that this is the case, call TimeRange.make_valid()

    In most cases, operations should treat end as exclusive.
    """

    def __init__(
        self,
        start: CrudeTimeStamp | TimeRange,
        end: CrudeTimeStamp | None = None,
        *args,
        **kwargs,
    ):
        if isinstance(start, TimeRange):
            start_ = cast(TimeRange, start)
            start, end = start_.start, start_.end
        end = start if end is None else end
        super().__init__(start=TimeStamp(start), end=TimeStamp(end))

    @classmethod
    def __get_validators__(cls):
        """Implemented for pydantic validation. Should not be used externally."""
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """Implemented for pydantic validation. Should not be used externally."""
        if isinstance(v, TimeRange):
            return cls(v.start, v.end)
        elif isinstance(v, dict):
            return cls(**v)
        return cls(start=v[0], end=v[1])

    @property
    def span(self) -> TimeSpan:
        """:returns: the TimeSpan between the start and end TimeStamps of the TimeRange."""
        return TimeSpan(self.end - self.start)

    def make_valid(self) -> TimeRange:
        """:returns: A copy of the Time Range with its start and end swapped if the
        TimeRange is invalid (i.e. its start is after its end). Otherwise, returns the
        TimeRange itself.
        """
        return self.swap() if not self.valid else self

    def clamp(self, bound: TimeRange) -> TimeRange:
        """Clamps a copy of the TimeRange and returns it, restricting its start and end
        to within the given bound TimeRange. The clamping is performed in a manner such
        that bound.contains(clamped) is guaranteed to return True.

        :param bound: The TimeRange to clamp by.
        :returns: The clamped TimeRange.
        """
        copy = self.copy()
        if bound.start.after(self.start):
            copy.start = bound.start
        if bound.start.after(self.end):
            copy.end = bound.start
        if bound.end.before(self.end):
            copy.end = bound.end
        if bound.end.before(self.start):
            copy.start = bound.end
        return copy

    def contains(self, value: TimeRange | CrudeTimeStamp) -> bool:
        """Checks if the TimeRange contains the given TimeRange or TimeStamp. If
        value is a TimeRange, it is considered contained if value's start is after or
        equal to the TimeRange start and value's end is before or equal to the TimeRange
        end. This means that two equal time ranges contain each other.

        If the value is a TimeStamp, it is considered contained if value is after
        or equal to the TimeRange start and value is STRICTLY before the TimeRange end.

        :param value: The TimeRange or TimeStamp to check.
        """
        if isinstance(value, TimeRange):
            return value.start.after_eq(self.start) and value.end.before_eq(self.end)
        pv = TimeStamp(value)
        return pv.after_eq(self.start) and pv.before(self.end)

    def overlaps_with(self, tr: TimeRange) -> bool:
        """:returns: True if the time ranges overlap.
        :param tr: The time range to compare against.
        """
        return self == tr or self.contains(tr.end) or tr.contains(self)

    def swap(self) -> TimeRange:
        """:returns: A new TimeRange with the start and end values swapped. Note that
        this may make the TimeRange invalid.
        """
        self.copy()
        return TimeRange(start=self.end, end=self.start)

    def copy(self, *args, **kwargs) -> TimeRange:
        """:returns: A copy of the time range."""
        return TimeRange(start=self.start, end=self.end)

    @property
    def valid(self) -> bool:
        """:returns: True if the TimeRange start is before or equal to the TimeRange
        end.
        """
        return self.span >= TimeSpan.ZERO

    def __str__(self) -> str:
        return str(self.start) + " - " + str(self.end)

    def __eq__(self, rhs: object) -> bool:
        if not isinstance(rhs, TimeRange):
            return False
        return self.start == rhs.start and self.end == rhs.end

    # Assigning these values to None prevents pydantic
    # from throwing a nasty missing attribute error.

    MAX: ClassVar[TimeRange] = None  # type: ignore
    """The maximum possible TimeRange, starting at TimeStamp.MIN and ending
    a TimeStamp.MAX
    """
    MIN: ClassVar[TimeRange] = None  # type: ignore
    """The minimum possible TimeRange, starting at TimeStamp.MAX and ending at
    TimeStamp.MIN. Note that this TimeRange is invalid.
    """
    ZERO: ClassVar[TimeRange] = None  # type: ignore
    """A TimeRange starting and ending at the unix epoch in UTC (TimeStamp.ZERO)."""


TimeRange.MAX = TimeRange(TimeStamp.MIN, TimeStamp.MAX)
TimeRange.MIN = TimeRange(TimeStamp.MAX, TimeStamp.MIN)
TimeRange.ZERO = TimeRange(0, 0)


class Density(int):
    """Density is the number of bytes contained in a single sample."""

    def __new__(cls, value: CrudeDensity):
        if isinstance(value, Density):
            return value
        if isinstance(value, int):
            return super().__new__(cls, value)
        raise TypeError(f"Cannot convert {type(value)} to Density")

    @classmethod
    def __get_validators__(cls):
        """Implemented for pydantic validation. Should not be used externally."""
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """Implemented for pydantic validation. Should not be used externally."""
        return cls(v)

    def sample_count(self, size: CrudeSize) -> int:
        """:returns: The number of contained in the given byte Size."""
        return int(size / self)

    def size_span(self, sample_count: int) -> Size:
        """:returns: The number of bytes occupied by the given number of samples."""
        return Size(sample_count * self)

    def __repr__(self):
        return f"Density({super().__repr__()})"

    UNKNOWN: Density
    """An unknown density, represented as a ZERO value."""
    BIT128: Density
    """A density of 128 bits or 16 bytes per sample."""
    BIT64: Density
    """A density of 64 bits or 8 bytes per sample."""
    BIT32: Density
    """A density of 32 bits or 4 bytes per sample."""
    BIT16: Density
    """A density of 16 bits of 2 bytes per sample."""
    BIT8: Density
    """A density of 8 bits or 1 bytes per sample."""


Density.UNKNOWN = Density(0)
Density.BIT128 = Density(16)
Density.BIT64 = Density(8)
Density.BIT32 = Density(4)
Density.BIT16 = Density(2)
Density.BIT8 = Density(1)


class Size(int):
    def __mul__(self, rhs: CrudeSize) -> Size:
        return Size(super(Size, self).__mul__(Size(rhs)))

    def __str__(self) -> str:
        tot_gb = self.trunc(Size.GB)
        tot_mb = self.trunc(Size.MB)
        tot_kb = self.trunc(Size.KB)
        tot_b = self.trunc(Size.BYTE)
        gb = tot_gb
        mb = tot_mb - tot_gb
        kb = tot_kb - tot_mb
        b = tot_b - tot_kb

        v = ""
        if gb != 0:
            v += f"{gb.gb_int}{Size.GB_UNITS} "
        if mb != 0:
            v += f"{mb.mb_int}{Size.MB_UNITS} "
        if kb != 0:
            v += f"{kb.kb_int}{Size.KB_UNITS} "
        if b != 0 or len(v) == 0:
            v += f"{b.bytes}{Size.BYTE_UNITS} "
        return v.strip()

    @property
    def gb(self) -> float:
        """:returns: The decimal number of gigabytes in the Size, including
        fractional gigabytes"""
        return float(self / Size.GB)

    @property
    def gb_int(self) -> int:
        """:returns: The integer number of gigabytes in the Size, NOT including
        fractional gigabytes"""
        return int(self / Size.GB)

    @property
    def mb(self) -> float:
        """:returns: The decimal number of megabytes in the Size, including
        fractional gigabytes"""
        return float(self / Size.MB)

    @property
    def mb_int(self) -> int:
        """:returns: The integer number of megabytes in the Size, NOT including
        fractional gigabytes"""
        return int(self / Size.MB)

    @property
    def kb(self) -> float:
        """:returns: The decimal number of kilobytes in the Size, including
        fractional gigabytes"""
        return float(self / Size.KB)

    @property
    def kb_int(self) -> int:
        """:returns: The integer number of kilobytes in the Size, NOT including
        fractional gigabytes"""
        return int(self / Size.GB)

    @property
    def bytes(self) -> int:
        """:returns: The number of bytes in the Size."""
        return int(self)

    def trunc(self, size: CrudeSize) -> Size:
        """Truncates the Size to th nearest integer of hte given size.

        For example,

        (Size.GIGABYTE + Size.MEGABYTE).trunc(Size.GIGABYTE)

        would return

        Size.GIGABYTE

        :param size: The Size to truncate by.
        """
        return Size(trunc(self / size) * size)

    def __sub__(self, rhs: CrudeSize) -> Size:
        return Size(int(self) - Size(rhs))

    def __add__(self, rhs: CrudeSize) -> Size:
        return Size(int(self) + Size(rhs))

    BYTE: Size
    BYTE_UNITS: SizeUnits
    KB: Size
    KB_UNITS: SizeUnits
    MB: Size
    MB_UNITS: SizeUnits
    GB: Size
    GB_UNITS: SizeUnits


Size.BYTE = Size(1)
Size.BYTE_UNITS = "b"
Size.KB = Size(1024) * Size.BYTE
Size.KB_UNITS = "kb"
Size.MB = Size(1024) * Size.KB
Size.MB_UNITS = "mb"
Size.GB = Size(1024) * Size.MB
Size.GB_UNITS = "gb"

SizeUnits = ["gb", "mb", "kb", "b"]


class DataType(str):
    """DataType represents a data type as a string."""

    def __new__(cls, value: CrudeDataType):
        if isinstance(value, DataType):
            return value

        if isinstance(value, str):
            return super().__new__(cls, value)

        if isinstance(value, np.number):
            value = DataType._FROM_NUMPY.get(np.dtype(value), None)
            if value is not None:
                return value

        if isinstance(value, float):
            return DataType.FLOAT64

        if isinstance(value, int):
            return DataType.INT64

        if isinstance(value, uuid.UUID):
            return DataType.UUID

        if isinstance(value, list):
            if len(value) == 0:
                raise ValueError("Cannot extract a data type from an empty list")

            if isinstance(value[0], TimeStamp):
                return DataType.TIMESTAMP

            if isinstance(value[0], float):
                return DataType.FLOAT64

            if isinstance(value[0], int):
                return DataType.INT64

            if isinstance(value[0], str):
                return DataType.STRING

            if isinstance(value[0], uuid.UUID):
                return DataType.UUID

            if isinstance(value[0], dict):
                return DataType.JSON

            raise TypeError(f"Cannot convert {type(value)} to DataType")

        if isinstance(value, dict):
            return DataType.JSON

        value = DataType._FROM_NUMPY.get(np.dtype(value), None)
        if value is not None:
            return value

        raise TypeError(f"Cannot convert {type(value)} to DataType")

    @classmethod
    def __get_validators__(cls):
        """Implemented for pydantic validation. Should not be used externally."""
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """Implemented for pydantic validation. Should not be used externally."""
        return cls(v)

    @classmethod
    def __modify_schema__(cls, field_schema):
        """Implemented for pydantic validation. Should not be used externally."""
        field_schema.update(type="string")

    @property
    def np(self) -> np.dtype:
        """Converts a built-in DataType to a numpy type Scalar Type
        :return: The numpy type
        """
        npt = DataType._TO_NUMPY.get(self, None)
        if npt is None:
            raise TypeError(f"Cannot convert {self} to numpy type")
        return cast(np.dtype, npt)

    @property
    def has_fixed_density(self) -> bool:
        """:returns: True if the DataType has a fixed density"""
        d = DataType._DENSITIES.get(self, None)
        return d is not None and d != Density.UNKNOWN

    @property
    def has_np(self) -> bool:
        """:returns: True if the DataType has a corresponding numpy type"""
        return self in DataType._TO_NUMPY

    @property
    def density(self) -> Density:
        """:returns: The density of the DataType. If the density can't be determined,
        returns Density.UNKNOWN.
        """
        return DataType._DENSITIES.get(self, Density.UNKNOWN)

    def __repr__(self):
        return f"DataType({super().__repr__()})"

    UNKNOWN: DataType
    UUID: DataType
    TIMESTAMP: DataType
    FLOAT64: DataType
    FLOAT32: DataType
    INT64: DataType
    INT32: DataType
    INT16: DataType
    INT8: DataType
    UINT64: DataType
    UINT32: DataType
    UINT16: DataType
    UINT8: DataType
    STRING: DataType
    JSON: DataType
    ALL: tuple[DataType, ...]
    _TO_NUMPY: dict[DataType, DTypeLike]
    _FROM_NUMPY: dict[DTypeLike, DataType]
    _DENSITIES: dict[DataType, Density]


DataType.UUID = DataType("uuid")
DataType.UNKNOWN = DataType("")
DataType.TIMESTAMP = DataType("timestamp")
DataType.FLOAT64 = DataType("float64")
DataType.FLOAT32 = DataType("float32")
DataType.INT64 = DataType("int64")
DataType.INT32 = DataType("int32")
DataType.INT16 = DataType("int16")
DataType.INT8 = DataType("int8")
DataType.UINT64 = DataType("uint64")
DataType.UINT32 = DataType("uint32")
DataType.UINT16 = DataType("uint16")
DataType.UINT8 = DataType("uint8")
DataType.JSON = DataType("json")
DataType.STRING = DataType("string")
DataType.ALL = (
    DataType.UUID,
    DataType.FLOAT64,
    DataType.FLOAT32,
    DataType.INT64,
    DataType.INT32,
    DataType.INT16,
    DataType.INT8,
    DataType.UINT64,
    DataType.UINT32,
    DataType.UINT16,
    DataType.UINT8,
    DataType.STRING,
    DataType.JSON,
)

CrudeTimeStamp: TypeAlias = (
    int | TimeStamp | TimeSpan | datetime | timedelta | np.datetime64 | np.int64
)
CrudeTimeSpan: TypeAlias = (
    int | float | TimeSpan | TimeStamp | timedelta | np.timedelta64
)
CrudeRate: TypeAlias = int | float | TimeSpan | Rate
CrudeDensity: TypeAlias = Density | int
CrudeDataType: TypeAlias = DTypeLike | DataType | str | list | np.number
CrudeSize: TypeAlias = int | float | Size

DataType._TO_NUMPY = {
    DataType.FLOAT64: np.dtype(np.float64),
    DataType.FLOAT32: np.dtype(np.float32),
    DataType.TIMESTAMP: np.dtype(np.int64),
    DataType.INT64: np.dtype(np.int64),
    DataType.INT32: np.dtype(np.int32),
    DataType.INT16: np.dtype(np.int16),
    DataType.INT8: np.dtype(np.int8),
    DataType.UINT64: np.dtype(np.uint64),
    DataType.UINT32: np.dtype(np.uint32),
    DataType.UINT16: np.dtype(np.uint16),
    DataType.UINT8: np.dtype(np.uint8),
}
DataType._FROM_NUMPY = {
    np.dtype(np.float64): DataType.FLOAT64,
    np.dtype(np.float32): DataType.FLOAT32,
    np.dtype(np.int64): DataType.INT64,
    np.dtype(np.int32): DataType.INT32,
    np.dtype(np.int16): DataType.INT16,
    np.dtype(np.int8): DataType.INT8,
    np.dtype(np.uint64): DataType.UINT64,
    np.dtype(np.uint32): DataType.UINT32,
    np.dtype(np.uint16): DataType.UINT16,
    np.dtype(np.uint8): DataType.UINT8,
    np.dtype(np.bool_): DataType.UINT8,
    np.dtype(np.datetime64): DataType.TIMESTAMP,
    np.dtype(np.str_): DataType.STRING,
}

DataType._DENSITIES = {
    DataType.UUID: Density.BIT128,
    DataType.FLOAT64: Density.BIT64,
    DataType.FLOAT32: Density.BIT32,
    DataType.TIMESTAMP: Density.BIT64,
    DataType.INT64: Density.BIT64,
    DataType.INT32: Density.BIT32,
    DataType.INT16: Density.BIT16,
    DataType.INT8: Density.BIT8,
    DataType.UINT64: Density.BIT64,
    DataType.UINT32: Density.BIT32,
    DataType.UINT16: Density.BIT16,
    DataType.UINT8: Density.BIT8,
    DataType.STRING: Density.UNKNOWN,
    DataType.JSON: Density.UNKNOWN,
}


class Bounds:
    lower: float
    upper: float

    def __init__(self, lower: float, upper: float):
        self.lower = lower
        self.upper = upper

    def contains(self, value: float) -> bool:
        return self.lower < value <= self.upper



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

import json
import uuid
from datetime import datetime

import numpy as np
import pandas as pd
from freighter import Payload
from pydantic import PrivateAttr

from synnax.telem.telem import (
    CrudeDataType,
    DataType,
    Size,
    TimeRange,
    TimeSpan,
    TimeStamp,
)
from synnax.util.interop import overload_comparison_operators


class Series(Payload):
    """Series is a strongly typed array of telemetry samples backed by an underlying
    binary buffer. It is interoperable with np.ndarray, meaning that it can be safely
    passed as an argument to any function/method that accepts a numpy array.

    Series also have an optional 'time_range' property that can be used to define the
    time range occupied by the Series' data. This property is guaranteed to be defined
    when reading data from a Synnax cluster, and is particularly useful for
    understanding the alignment of samples in relation to another series. This is
    especially relevant in the context of a Frame (framer.Frame).
    """

    time_range: TimeRange | None = None
    """An optional property defines the time range occupied by the
    Series' data. This property is guaranteed to be defined when reading data from
    a Synnax cluster, and is particularly useful for understanding the alignment of
    samples in relation to another series. This is especially relevant in the context
    of a Frame (framer.Frame). When set by the Cluster, the start of the time range
    represents the timestamp of the first sample in the array (inclusive), while the
    end of the time range is set to the nanosecond AFTER the last sample in the array
    (exclusive).
    """
    data_type: DataType
    """The data type of the Series"""
    data: bytes
    """The underlying buffer"""
    alignment: int = 0
    __len_cache: int | None = PrivateAttr(None)

    def __len__(self) -> int:
        if self.data_type.has_fixed_density:
            return self.data_type.density.sample_count(len(self.data))
        if self.__len_cache is None:
            self.__len_cache = self.data.count(b"\n")
        return self.__len_cache

    def __init__(
        self,
        data: CrudeSeries,
        data_type: CrudeDataType | None = None,
        time_range: TimeRange | None = None,
        alignment: int = 0,
    ):
        if isinstance(data, (TimeStamp, int, float, np.number)):
            data_type = data_type or DataType(data)
            data_ = np.array([data], dtype=data_type.np).tobytes()
        elif isinstance(data, Series):
            data_type = data_type or data.data_type
            data_ = data.data
            time_range = data.time_range if time_range is None else time_range
        elif isinstance(data, MultiSeries):
            if len(data.series) == 1:
                data_type = data_type or data.series[0].data_type
                data_ = data.series[0].data
                time_range = (
                    data.series[0].time_range if time_range is None else time_range
                )
            else:
                raise ValueError(
                    "[Series] - MultiSeries with more than one series cannot be converted to a Series"
                )
            data_type = data_type or data
        elif isinstance(data, pd.Series):
            data_type = data_type or DataType(data.dtype)
            data_ = data.to_numpy(dtype=data_type.np).tobytes()
        elif isinstance(data, np.ndarray):
            data_type = data_type or DataType(data.dtype)
            data_ = data.astype(data_type.np).tobytes()
        elif isinstance(data, list):
            data_type = data_type or DataType(data)
            if data_type == DataType.JSON:
                data_ = (
                    b"\n".join([json.dumps(d).encode("utf-8") for d in data]) + b"\n"
                )
            elif data_type == DataType.STRING:
                data_ = b"\n".join([d.encode("utf-8") for d in data]) + b"\n"
            elif data_type == DataType.UUID:
                data_ = b"".join(d.bytes for d in data)
            else:
                data_ = np.array(data, dtype=data_type.np).tobytes()
                data_type = data_type or DataType(data)
        elif isinstance(data, str):
            data_ = bytes(f"{data}\n", "utf-8")
            data_type = DataType.STRING
        else:
            if data_type is None:
                raise ValueError(
                    "[Series] - data_type must be specified if a buffer is given",
                )
            data_type = DataType(data_type)
            data_ = data
        super().__init__(
            data_type=data_type, data=data_, time_range=time_range, alignment=alignment
        )
        self.__len_cache = None

    class Config:
        arbitrary_types_allowed = True

    def __array__(self, *args, **kwargs) -> np.ndarray:
        """Implemented to that the Series can be passed around as a numpy array. See
        https://numpy.org/doc/stable/user/basics.interoperability.html#the-array-method.
        """
        if not self.data_type.has_np:
            raise ValueError(
                f"""
                [Series] - {self.data_type} does not have a numpy equivalent, so it can't
                be interpreted as a numpy array.
                """
            )

        if len(args) > 0:
            return np.array(self.__array__(), *args, **kwargs)
        return np.frombuffer(self.data, dtype=self.data_type.np, **kwargs)

    def to_numpy(self, *args, **kwargs) -> np.ndarray:
        """Converts the Series to a numpy array. This is necessary for matplotlib
        interop.
        """
        return self.__array__(*args, **kwargs)

    def __getitem__(self, index: int) -> SampleValue:
        if not self.data_type.has_np and index < 0:
            index = len(self) + index

        if self.data_type == DataType.UUID:
            start = self.data_type.density.size_span(index)
            end = start + self.data_type.density
            d = self.data[start:end]
            return uuid.UUID(bytes=d)

        if self.data_type == DataType.JSON:
            d = self.__newline_getitem__(index)
            return json.loads(d)

        if self.data_type == DataType.STRING:
            d = self.__newline_getitem__(index)
            return d.decode("utf-8")

        return self.__array__()[index]

    def __newline_getitem__(self, index: int) -> bytes:
        if index == 0:
            start = 0
        else:
            start = self.data.find(b"\n")
            while start >= 0 and index > 1:
                start = self.data.find(b"\n", start + 1)
                index -= 1
            start += 1

        if start < 0:
            raise IndexError(f"[Series] - Index {index} out of bounds")

        end = self.data.find(b"\n", start)
        if end < 0:
            end = len(self.data)
        return self.data[start:end]

    def __iter__(self):
        if self.data_type == DataType.UUID:
            yield from [self[i] for i in range(len(self))]
        elif self.data_type == DataType.JSON:
            for v in self.__iter__newline():
                yield json.loads(v)
        elif self.data_type == DataType.STRING:
            for v in self.__iter__newline():
                yield v.decode("utf-8")
        else:
            yield from self.__array__()

    def __iter__newline(self):
        curr = 0
        while curr < len(self.data):
            end = self.data.find(b"\n", curr)
            if end < 0:
                end = len(self.data)
            yield self.data[curr:end]
            curr = end + 1

    def __str__(self) -> str:
        return str(list(self))

    @property
    def size(self) -> Size:
        """:returns: A Size representing the number of bytes in the Series' data."""
        return Size(len(self.data))

    def astype(self, data_type: DataType) -> Series:
        return Series(
            data=self.__array__().astype(data_type.np),
            data_type=data_type,
            time_range=self.time_range,
        )

    def to_datetime(self) -> list[datetime]:
        return [pd.Timestamp(t).to_pydatetime() for t in self.__array__()]

    def __eq__(self, other):
        if isinstance(other, Series):
            return self.data == other.data
        elif isinstance(other, np.ndarray):
            return self.__array__() == other
        else:
            return False


Series = overload_comparison_operators(Series, "__array__")

SampleValue = np.number | uuid.UUID | dict | str | int | float | TimeStamp
TypedCrudeSeries = Series | pd.Series | np.ndarray
CrudeSeries = (
    Series
    | bytes
    | pd.Series
    | np.ndarray
    | list[float]
    | list[str]
    | list[dict]
    | float
    | int
    | TimeStamp
)


def elapsed_seconds(d: np.ndarray) -> np.ndarray:
    """Converts a Series of timestamps to elapsed seconds since the first timestamp.

    :param d: A Series of timestamps.
    :returns: A Series of elapsed seconds.
    """
    return (d - d[0]) / TimeSpan.SECOND


class MultiSeries:
    series: list[Series]

    def __init__(self, series: list[Series]):
        self.series = series
        if len(series) > 0:
            first_dt = series[0].data_type
            same_dt = all(s.data_type == first_dt for s in series)
            if not same_dt:
                raise ValueError(
                    f"""
                    [MultiSeries] - All series must have the same data type. Received
                    {first_dt} and {set(s.data_type for s in series)}.
                    """
                )

    def __array__(self) -> np.ndarray:
        pre_alloc = np.empty((len(self),), dtype=self.series[0].data_type.np)
        start = 0
        for s in self.series:
            end = start + len(s)
            pre_alloc[start:end] = s.__array__()
            start = end
        return pre_alloc

    def to_numpy(self) -> np.ndarray:
        return self.__array__()

    @property
    def time_range(self) -> TimeRange | None:
        if len(self.series) == 0:
            return TimeRange.ZERO
        first = self.series[0].time_range
        last = self.series[-1].time_range
        if first is None or last is None:
            return None
        return TimeRange(start=first.start, end=last.end)

    def __len__(self) -> int:
        return sum(len(s) for s in self.series)

    def __getitem__(self, index: int) -> SampleValue:
        if index < 0:
            index = len(self) + index
        for s in self.series:
            if index < len(s):
                return s[index]
            index -= len(s)
        raise IndexError(f"[MultiSeries] - Index {index} out of bounds for {len(self)}")

    def __iter__(self):
        for s in self.series:
            yield from s

    @property
    def size(self) -> Size:
        return Size(sum(s.size for s in self.series))


MultiSeries = overload_comparison_operators(MultiSeries, "__array__")



#  Copyright 2024 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from uuid import UUID
from freighter import Payload
from synnax.ontology.payload import ID


class NewUser(Payload):
    username: str
    password: str
    first_name: str = ""
    last_name: str = ""
    key: UUID | None = None


class User(Payload):
    key: UUID
    username: str
    first_name: str
    last_name: str
    root_user: bool

    def ontology_id(self) -> ID:
        return ID(key=self.key, type="user")


ontology_type = "user"
change_username_action = "change_username"
